//==============================================================================
//	$Id: pdfmaker.js,v 1.32 2019/08/29 15:29:04 wolf Exp $
//==============================================================================
//
//	OpenLayers.Layer.Vector.CAOS
//
//==============================================================================

var AFM = window.AFM || {};

var PDFMAKER = function() {

	//======================================================================
	//	PDF name
	//======================================================================

	this.name = function (name) {

		if (name.substr(0, 1) == '\x0e') return name;
		return '\x0e' + name;
	};

	//======================================================================
	//	PDF objects
	//======================================================================

	this.objects = [null];

	this.createObject = function (type) {

		var id = this.objects.length;

		var object = {'#': id};
		this.objects.push(object);

		if (typeof(type)=='string') {

			object.Type = this.name(type);

		} else if (type) {

			for (var key in type) {
				object[key] = type[key];
			}
		}

		return object;
	};

	//======================================================================
	//	PDF Basisobjekte
	//======================================================================

	this.info = this.createObject('Info');
	this.main = this.createObject('Catalog');
	this.root = this.createObject('Pages');

	this.main.Version = this.name('1.4');
	this.main.Pages   = this.root;
	this.root.Count   = 0;
	this.root.Kids    = [];
	this.root.Resources = {};
	//this.root.Resources.ProcSet = [this.name('PDF'), this.name('Text')];

	//======================================================================
	//	Layout
	//======================================================================

	this.mediaWidth   = 595.2756;
	this.mediaHeight  = 841.8898;

	this.borderLeft   = 36;
	this.borderBottom = 72;
	this.borderRight  = 36;
	this.borderTop    = 72;

	this.setContentSize = function () {

		this.contentX     = this.borderLeft;
		this.contentY     = this.mediaHeight - this.borderTop;
		this.contentWidth = this.mediaWidth  - this.borderLeft - this.borderRight;
		this.contentHeight= this.mediaHeight - this.borderTop  - this.borderBottom;
	};

	this.setContentSize();

	this.setMediaSize = function (width, height) {

		this.mediaWidth  = width;
		this.mediaHeight = height;
		this.root.MediaBox = [0, 0, width, height];

		this.setContentSize();
	};

	this.setMargins = function (left, bottom, right, top) {

		this.borderLeft   = left;
		this.borderBottom = bottom;
		this.borderRight  = right;
		this.borderTop    = top;

		this.setContentSize();
	};

	//======================================================================
	//	Document info
	//======================================================================

	this.setCreator = function (creator) {

		this.info.Creator = '' + creator;
	};

	this.setAuthor = function (author) {

		this.info.Author = '' + author;
	};

	this.setSubject = function (subject) {

		this.info.Subject = '' + subject;
	};

	this.setTitle = function (title) {

		this.info.Title = '' + title;
	};

	this.setKeywords = function (keywords) {

		this.info.Keywords = '' + keywords;
	};

	//======================================================================
	//	Pagetree
	//======================================================================

	this.createPageDirectory = function (parent) {

		var pagedir = this.createObject ('Pages');

		pagedir.Parent = parent || this.root;
		pagedir.Parent.Count = pagedir.Parent.Kids.length;
		pagedir.Parent.Kids.push(pagedir);

		pagedir.Count = 0;
		pagedir.Kids = [];

		return pagedir;
	};

	this.createPage = function (parent) {

		parent = parent || this.root;

		var page = this.createObject ('Page');

		page.Parent = parent || this.root;
		page.Parent.Kids.push(page);
		page.Parent.Count = page.Parent.Kids.length;
		page._pdf = this;
		page._y   = this.contentY;
		page._x   = this.contentX;
		page._createTextBox= function(sname, fontSize, opts) {
			return this._pdf.createTextBox(sname, fontSize, opts);
		};

		return page;
	};

	//======================================================================
	//	Stream
	//======================================================================

	this.createWriterForPage = function (page) {

		var contentArray = [];

		var stream = this.createObject (null);
		stream.Length = 0;
		stream['!'] = contentArray;
		stream['/'] = '\n';

		if (!page.Contents) page.Contents = [];
		page.Contents.push(stream);

		function writer () {

			for (var i in arguments) {

				var argument = arguments[i];

				if (argument instanceof Array) {

					for (var j in argument) {

						writer(argument[j]);
					}
					continue;
				}

				contentArray.push(argument);
			}
		}

		return writer;
	};

	//======================================================================
	//	XObjects
	//======================================================================

	this.useXObject = function (sname, xobject, pageOrPages) {

		pageOrPages = pageOrPages || this.root;
		if (!pageOrPages.Resources) pageOrPages.Resources = {};
		if (!pageOrPages.Resources.XObject) pageOrPages.Resources.XObject = {};
		pageOrPages.Resources.XObject[sname.substr(1)] = xobject;
	};

	//======================================================================
	//	Colors
	//======================================================================

	this.gray = function(gray) {

		return {
			f: gray + ' g',
			s: gray + ' G'
		};
	};

	this.WHITE	= this.gray (1);
	this.GRAY	= this.gray (0.5);
	this.LIGHTGRAY	= this.gray (0.75);
	this.BLACK	= this.gray (0);

	this.rgb = function(r, g, b) {

		return {
			f: r + ' ' + g + ' ' + b + ' rg',
			s: r + ' ' + g + ' ' + b + ' RG'
		};
	};

	this.BLUE   = this.rgb (0, 0, 1);
	this.GREEN  = this.rgb (0, 1, 0);
	this.RED    = this.rgb (1, 0, 0);
	this.YELLOW = this.rgb (1, 1, 0);
	this.CYAN   = this.rgb (0, 1, 1);
	this.MAGENTA= this.rgb (1, 0, 1);

	//======================================================================
	//	Fonts
	//======================================================================

	this.nFonts = 0;
	this.fontDictionaryBySname = {};
	this.fonts = {};
	this.fontMetricBySname = {};

	this.useFont = function (sname, pageOrPages) {

		if (!pageOrPages.Resources) pageOrPages.Resources = {};
		if (!pageOrPages.Resources.Font) pageOrPages.Resources.Font = {};
		pageOrPages.Resources.Font[sname.substr(1)] = this.fontDictionaryBySname[sname];
	};

	this.getFont = function (baseFont, pageOrPages, encoding) {

		pageOrPages = pageOrPages || this.root;
		encoding = encoding || 'WinAnsiEncoding';

		if (!this.fonts[baseFont]) this.fonts[baseFont] = {};

		var sname = this.fonts[baseFont][encoding];

		if (sname) {

			this.useFont(sname, pageOrPages);
			return sname;
		}

		//--------------------------------------------------------------
		//	create new font
		//--------------------------------------------------------------

		var metric = AFM[baseFont];
		if (!metric) {

			window.alert ('Fontmetric for "' + baseFont + '" not available.');
			return null;
		}

		++this.nFonts;
		var name = 'F' + this.nFonts;

		var font = this.createObject ('Font');
		font.BaseFont = this.name(baseFont);
		font.Encoding = this.name(encoding);
		font.Name     = this.name(name);
		font.Subtype  = this.name('Type1');

		sname = '/' +name;

		this.fontDictionaryBySname[sname] = font;
		this.fonts[baseFont][encoding] = sname;
		this.fontMetricBySname[sname] = metric;

		this.useFont(sname, pageOrPages);

		return sname;
	};

	this.unknownCharCodes = {};

	this.getWidth = function (metric, s, start, length) {

		//--------------------------------------------------------------
		//	prepare slice
		//--------------------------------------------------------------

		if (length===0) return 0;

		start = start || 0;
		var end   = length? Math.min (s.length, start+length) : s.length;

		if (end<=start) return 0;

		//--------------------------------------------------------------
		//	prepare metric
		//--------------------------------------------------------------

		if (typeof(metric)=='string') {

			metric = this.fontMetricBySname[metric];
		}

		var charMetrics = metric.CharMetrics;

		//--------------------------------------------------------------
		//	measure width
		//--------------------------------------------------------------

		var width = 0;

		for (var index=start; index<end; index++) {

			var ch = s.charAt(index);
			var cm = charMetrics[ch];

			if (!cm) {

				if (!this.unknownCharCodes[ch]) {

					this.unknownCharCodes[ch] = true;
					window.alert ('No metric for char code ' +
						ch.charCodeAt(0) + '.');
				}
				continue;
			}

			width += cm.w;
		}

		return width;
	};

	//======================================================================
	//	Formatierer
	//======================================================================

	this.dd = function (d) {

		if (d<10) return '0' + d; else return '' + d;
	};

	this.dl = function (d, l) {

		d = '' + d;
		while (d.length < l) d = '0' + d;
		return d;
	};

	this.pdftime = function (date) {

		var offset = -date.getTimezoneOffset();

		var mins = offset % 60;
		var hours = (offset - mins) / 60;

		return'D:' +
			date.getFullYear() +
			this.dd(date.getMonth()+1) +
			this.dd(date.getDate()) +
			this.dd(date.getHours()) +
			this.dd(date.getMinutes()) +
			this.dd(date.getSeconds()) +
			(offset<0? '-' : '+') +
			this.dd(hours) + '\x27' +
			(mins? this.dd(mins) + '\x27' : '');
	};

	this.pdfString = function (s) {

		//--------------------------------------------------------------
		//	String ohne Spezialzeichen
		//--------------------------------------------------------------

		if (s.match (/^[^()\\]*$/)) return '(' + s + ')';

		//--------------------------------------------------------------
		//	String mit Escapes
		//--------------------------------------------------------------

		var result = ['('];
		var parts = s.match (/[^()\\]+|[()\\]/g);

		for (var i in parts) {

			var part = parts[i];

			switch (part.substr(0,1)) {

			case '(':
			case ')':
			case '\\':
				result.push('\\');
				break;

			default:
				break;
			}

			result.push(part);
		}

		result.push(')');
		return result.join('');
	};

	//======================================================================
	//
	//	Displaybox
	//
	//======================================================================

	this.createBox = function (opts) {

		opts = opts || {};

		var box = {

			pdf:		this,

			minWidth:	opts.minWidth	|| opts.width || 0,
			maxWidth:	opts.maxWidth	|| opts.width || Infinity,

			minHeight:	opts.minHeight	|| opts.height || 0,
			maxHeight:	opts.maxHeight	|| opts.height || Infinity,

			contentWidth:	0,
			contentHeight:	0,

			gravity:	opts.gravity	|| 'tl',

			bgColor:	opts.bgColor	|| null,
			borderColor:	opts.borderColor|| null,
			borderWidth:	opts.borderWidth|| 0,

			r: function(v,d) { return v.toFixed(d)*1; }
		};

		box.width  = box.minWidth;
		box.height = box.minHeight;

		box.getExtent = function() {

			this.width  = this.r(Math.min (Math.max (this.contentWidth,  this.minWidth ), this.maxWidth ), 3);
			this.height = this.r(Math.min (Math.max (this.contentHeight, this.minHeight), this.maxHeight), 3);

			return {width: this.width, height: this.height};
		};

		box.setMinExtent = function (newMinWidth, newMinHeight) {

			if (newMinWidth && newMinWidth > this.minWidth) {

				this.minWidth = Math.min (newMinWidth, this.maxWidth);
			}

			if (newMinHeight && newMinHeight > this.minHeight) {

				this.minHeight = Math.min (newMinHeight, this.maxHeight);
			}

			this.getExtent();
		};

		box.getBackground = function() {

			if (!this.bgColor && !this.borderColor) return [];

			var result = [];

			result.push ('0 0 ' + this.width + ' ' + -this.height + ' re');

			if (this.bgColor) result.push (this.bgColor.f);
			if (this.borderColor) {

				result.push (this.borderColor.s);
				result.push (this.borderWidth + ' w');
				result.push (this.bgColor? 'b' : 's');

			} else if (this.bgColor) {

				result.push ('f');
			}

			return result.join(' ');
		};

		box.getResult = function() {

			this.getExtent();

			var result = this.getBackground();
			return result.length ? ['q', result, 'Q'] : result;
		};

		return box;
	};

	//======================================================================
	//	TableBox
	//======================================================================

	this.createTableBox = function (opts) {

		opts = opts || {};

		var box = this.createBox (opts);

		box.cellSpacing	= opts.cellSpacing || 0;
		box.type = 'table';

		box.colWidths= [];
		box.headRows = [];
		box.bodyRows = [];
		box.footRows = [];

		box.setColWidths = function(row) {

			for (var i in row.fields) {

				var field = row.fields[i];
				this.colWidths[i] = Math.max (field.width, (this.colWidths[i] || 0));
			}
		};

		box.appendHeadRow = function (child) {

			child.getExtent();

			if (this.headRows.length &&
				this.contentHeight + this.cellSpacing + child.height > this.maxHeight) {

				return false;
			}

			this.headRows.push(child);
			this.setColWidths(child);

			if (this.contentHeight) this.contentHeight += this.cellSpacing;
			this.contentHeight += child.height;
			return true;
		};

		box.appendBodyRow = function (child) {

			child.getExtent();

			if (this.bodyRows.length &&
				this.contentHeight + this.cellSpacing + child.height > this.maxHeight) {

				return false;
			}

			this.bodyRows.push(child);
			this.setColWidths(child);

			if (this.contentHeight) this.contentHeight += this.cellSpacing;
			this.contentHeight += child.height;
			return true;
		};

		box.appendFootRow = function (child) {

			child.getExtent();

			if (this.footRows.length &&
				this.contentHeight + this.cellSpacing + child.height > this.maxHeight) {

				return false;
			}

			this.footRows.push(child);
			this.setColWidths(child);

			if (this.contentHeight) this.contentHeight += this.cellSpacing;
			this.contentHeight += child.height;
			return true;
		};

		box.getResult = function() {

			this.contentWidth = Math.max(0, this.colWidths.length-1) * this.cellSpacing;

			for (var i in this.colWidths) {

				this.contentWidth += this.colWidths[i];
			}

			this.getExtent();

			var result = [];
			result.push ('q');
			result.push (this.getBackground());

			var offset = 0;

			for (i in this.headRows) {

				if (offset) result.push ('1 0 0 1 0 ' + -offset + ' cm');

				var row = this.headRows[i];
				row.getExtent();
				result.push (row.getResult(this));
				offset = row.height + this.cellSpacing;
			}

			for (i in this.bodyRows) {

				if (offset) result.push ('1 0 0 1 0 ' + -offset + ' cm');

				row = this.bodyRows[i];
				row.getExtent();
				result.push (row.getResult(this));
				offset = row.height + this.cellSpacing;
			}

			for (i in this.footRows) {

				if (offset) result.push ('1 0 0 1 0 ' + -offset + ' cm');

				row = this.footRows[i];
				row.getExtent();
				result.push (row.getResult(this));
				offset = row.height + this.cellSpacing;
			}

			result.push ('Q');
			return result;
		};

		return box;
	};

	//======================================================================
	//	RowBox
	//======================================================================

	this.createRowBox = function (opts) {

		opts = opts || {};

		var box = this.createBox (opts);

		box.type = 'row';
		box.fields = [];

		box.appendField = function (child) {

			child.getExtent();

			this.contentWidth += child.width;
			this.contentHeight = Math.max(this.contentHeight, child.height);
			this.fields.push(child);
		};

		box.getResult = function (table) {

			table = table || {};
			var fieldWidths = table.colWidths || [];
			var cellSpacing = table.cellSpacing || 0;

			this.minWidth = Math.max (this.minWidth, (table.width||0));
			this.getExtent();

			var result = [];
			result.push ('q');
			result.push (this.getBackground());

			var offset = 0;

			for (var i in this.fields) {

				if (offset) result.push ('1 0 0 1 ' + offset + ' 0 cm');

				var field = this.fields[i];

				field.minWidth = Math.max (field.minWidth,  fieldWidths[i]||0);
				field.minHeight= Math.max (field.minHeight, this.height);

				result.push (field.getResult());
				offset = field.width + cellSpacing;
			}

			result.push ('Q');

			return result;
		};

		return box;
	};

	//======================================================================
	//	ImageBox
	//======================================================================

	this.nImages = 0;

	this.createImageBox = function (imageObject, opts) {

		if (!imageObject['#']) imageObject = this.createObject(imageObject);

		opts = opts || {};

		var box = this.createBox (opts);

		var scale = opts.scale || 1.0;
		var w = scale * imageObject.Width;
		var h = scale * imageObject.Height;

		var s = Math.min (box.maxWidth/w, box.maxHeight/h, 1);

		box.contentWidth = box.r(s * w, 2);
		box.contentHeight= box.r(s * h, 2);

		++this.nImages;
		box.sname = '/Img' + this.nImages;

		this.useXObject (box.sname, imageObject);

		box.getResult = function () {

			this.getExtent();

			return [
				this.getBackground(), [
				'q',
				this.contentWidth, 0, 0, this.contentHeight,
				(this.width   - this.contentWidth) / 2,
				-(this.height + this.contentHeight) / 2,
				'cm',
				this.sname, 'Do',
				'Q'].join(' ')
			];
		};

		return box;
	};

	//======================================================================
	//	Textbox
	//======================================================================

	this.createTextBox = function (sname, fontSize, opts) {

		opts = opts || {};

		var box = this.createBox (opts);

		box.lrAlign	= opts.lrAlign	|| 'l';
		box.tbAlign	= opts.tbAlign	|| 't';
		box.justify	= opts.justify	|| 0;
		box.textPadding	= opts.padding	|| 0;
		box.textColor	= opts.textColor|| this.BLACK;

		//--------------------------------------------------------------
		//	context
		//--------------------------------------------------------------

		box.pdf		= this;
		box.metric	= this.fontMetricBySname[sname];
		box.fontSize	= fontSize;

		//--------------------------------------------------------------
		//	result
		//--------------------------------------------------------------

		box.result	= [];
		box.lineHeight	= box.r(fontSize * (opts.lineHeight||1.2), 1);
		box.contentHeight=2*box.textPadding;

		//--------------------------------------------------------------
		//	internal
		//--------------------------------------------------------------

		box.u2g = function(userValue)  { return userValue /this.fontSize*1000; };
		box.g2u = function(glyphValue) { return glyphValue*this.fontSize/1000; };

		var baselineY	= (1000 + box.metric.Ascender + box.metric.Descender) / 2000;

		box.gWidth	= 0;
		box.gMaxWidth	= box.u2g(box.maxWidth-2*box.textPadding);
		box.gSpaceWidth	= box.metric.CharMetrics['\x20'].w;
		box.words	= [];
		box.op		= 'Tj';

		//--------------------------------------------------------------
		//	prefix
		//--------------------------------------------------------------

		box.result.push (sname + ' ' + fontSize + ' Tf');
		box.result.push (0 + ' ' + box.r(-box.lineHeight*baselineY, 1) + ' Td');
		box.result.push (box.lineHeight + ' TL');

		//--------------------------------------------------------------
		//	workhorse
		//--------------------------------------------------------------

		box.indent = function (newIndent) {

			var indent = newIndent - (this.currentIndent || 0);
			if (!indent) return;

			this.currentIndent = newIndent;
			this.result.push (indent + ' 0 Td');
		};

		box.nextLine = function (inblock) {

			this.contentWidth   = Math.max (this.contentWidth, this.g2u(this.gWidth)+2*this.textPadding);
			this.contentHeight += this.lineHeight;

			if (!this.words.length) {

				this.result.push ('T*');
				return;
			}

			var string = this.pdf.pdfString(this.words.join('\x20'));

			switch (this.lrAlign) {

			case 'l':
				break;

			case 'r':
				this.indent(-this.g2u(this.gWidth));
				break;

			case 'c':
				this.indent(-this.g2u(this.gWidth)/2);
				break;

			case 'j':
				if (this.gMaxWidth >= Infinity) break;

				var blankSpace = Math.max (this.g2u(this.gMaxWidth-this.gWidth), 0);

				if (this.justify<1) {

					var units = inblock && string.length>=2?
						blankSpace/string.length : 0;

					this.result.push (this.r(units, 3) + ' Tc');

				} else {

					units = inblock && this.words.length>=2?
					 	blankSpace / (this.words.length-1) : 0;

					this.result.push (this.r(units, 3) + ' Tw');
				}
				break;
			default:
				break;
			}

			this.result.push (string + this.op);
			this.op = '\'';
			this.words = [];
			this.gWidth = 0;
		};

		box.flush = function () {

			if (this.words.length) this.nextLine();
		};

		box.write = function (s) {

			if (!s) return;

			var lines = s.split('\n');

			for (var i=0; i<lines.length; i++) {

				if (i>0) this.writeln();
				this.writeWords(lines[i]);
			}
		};

		box.writeWords = function (s) {

			if (!s) return;

			var words = s.replace(/[^\x20-\xff]/g, '\xbf').split(' ');

			for (var i in words) {

				this.writeWord(words[i]);
			}
		};

		box.writeWord = function (word) {

			if (!word.length) return;

			var availableWidth = this.gMaxWidth - this.gWidth -
				(this.words.length ? this.gSpaceWidth : 0);

			//------------------------------------------------------
			//	ganzes Wort passt
			//------------------------------------------------------

			var noShyWord   = word.replace(/\xad/g,'');
			var gNoShyWidth = this.pdf.getWidth(this.metric, noShyWord);

			if (gNoShyWidth <= availableWidth) {

				if (this.gWidth) this.gWidth += this.gSpaceWidth;
				this.gWidth += gNoShyWidth;
				this.words.push (noShyWord);
				return;
			}

			//------------------------------------------------------
			//	nicht trennbares Wort --> auf neue Zeile
			//------------------------------------------------------

			var parts = word.split(/([\xad-])/);

			if (parts.length <= 1) {

				if (gNoShyWidth > this.gMaxWidth) {
					window.console.log ('Word to long: “' + word + '”');
				}

				if (this.gWidth) this.nextLine(true);
				this.gWidth += gNoShyWidth;
				this.words.push (noShyWord);
				return;
			}

			//------------------------------------------------------
			//	Umbruch wenn erster Wortteil nicht passt
			//------------------------------------------------------

			if (this.gWidth &&
				this.pdf.getWidth(this.metric, parts[0]+parts[1]) > availableWidth) {

				this.nextLine(true);
				availableWidth = this.gMaxWidth;
			}

			//------------------------------------------------------
			//	Hyphanate word
			//------------------------------------------------------

			var fitWord = parts[0];
			var fitSepa = parts[1] || '';
			var fitWidth= this.pdf.getWidth(this.metric, fitWord+fitSepa);

			var index;

			for (index=2; index < parts.length; index+=2) {

				var nextPart = parts[index];
				var nextSepa = parts[index+1] || '';

				var fitHyphen = fitSepa == '-' ? '-' : '';

				var testWord = fitWord + fitHyphen + nextPart + nextSepa;
				var testWidth= this.pdf.getWidth(this.metric, testWord);

				if (testWidth > availableWidth) break;

				fitWord = fitWord + fitHyphen + nextPart;
				fitSepa = nextSepa;
				fitWidth= testWidth;
			}

			if (this.gWidth) this.gWidth += this.gSpaceWidth;
			this.gWidth += fitWidth;
			this.words.push (fitWord+fitSepa);

			var remainder = parts.slice(index).join('');
			if (remainder) {
				this.nextLine(true);
				this.writeWord(remainder);
			}
		};

		box.writeln = function (s) {

			if (s) this.write(s);
			this.nextLine();
		};

		//--------------------------------------------------------------
		//	result
		//--------------------------------------------------------------

		box.getResult = function () {

			this.flush();
			this.getExtent();

			var offsetX = +this.textPadding;
			var offsetY = -this.textPadding;

			if (this.contentHeight > this.gMaxHeight) {
				window.console.log ('Textbox to high');
			}

			switch(box.tbAlign) {

			case 'm':
				offsetY -= (this.minHeight - this.contentHeight) / 2;
				break;

			case 'b':
				offsetY -= (this.minHeight - this.contentHeight);
				break;

			default:
				break;
			}

			switch (this.lrAlign) {

			case 'c':
				offsetX = this.width/2;
				break;

			case 'r':
				offsetX = this.width - this.textPadding;
				break;

			default:
				break;
			}

			var result = [];
			result.push (this.getBackground());
			result.push (this.textColor.f);
			result.push ('BT');

			if (offsetX || offsetY) {

				result.push (offsetX + ' ' + offsetY + ' Td');
			}

			result.push (this.result);
			result.push ('ET');

			return result;
		};

		if (opts.text) {

			box.write(opts.text);
			box.flush();
			box.getExtent();
		}

		return box;
	};

	//======================================================================
	//	Renderer
	//======================================================================

	this.objectToString = function(object, indent) {

		var myindent = indent || '';

		var type = typeof(object);
		switch (type) {

		case 'boolean':
		case 'number':

			return '' + object;

		case 'string':

			//------------------------------------------------------
			//	PDF name
			//------------------------------------------------------

			if (object.substr(0,1) == '\x0e') {

				return '/' + object.substr(1);
			}

			//------------------------------------------------------
			//	PDF string
			//------------------------------------------------------

			return this.pdfString (object);

		case 'object':

			var subIndent = myindent + '\t';

			//------------------------------------------------------
			//	Array
			//------------------------------------------------------

			if (object instanceof Array) {

				var elements = [];

				for (var i in object) {

					elements.push (this.objectToString(object[i], subIndent));
				}

				return '[' + elements.join(' ') + ']';
			}

			//------------------------------------------------------
			//	PDF object reference
			//------------------------------------------------------

			if (object['#']) return object['#'] + ' 0 R';

			//------------------------------------------------------
			//	PDF dictionary
			//------------------------------------------------------

			var components = [];

			components.push ('<<');

			for (name in object) {

				if (name < 'A' || name.substring(0, 1) == '_') continue;
				components.push('/' + name + ' ' +
					this.objectToString(object[name], subIndent));
			}

			return components.join('\n' + subIndent) + '\n' + myindent + '>>';
		default:
			break;
		}

		return '{{' + object + ' : ' + type + '}}';
	};

	this.toString = function() {

		//--------------------------------------------------------------
		//	ergaenze Info-Objekt
		//--------------------------------------------------------------

		this.info.CreationDate	= this.pdftime(new Date());
		this.info.ModDate	= this.info.CreationDate;
		this.info.Producer	= 'pdfmaker.js\xa9netzwolf.info';

		//--------------------------------------------------------------
		//	PDF Kopf
		//--------------------------------------------------------------

		var components = [];
		var byteOffset = 0;

		var pdfIntro = '%PDF-1.4\n%\x93\x8c\x8b\x93\n\n';

		components.push (pdfIntro);
		byteOffset += pdfIntro.length;

		//--------------------------------------------------------------
		//	PDF Objekte
		//--------------------------------------------------------------

		var objOffsets = [0];

		for (var objNum=1; objNum<this.objects.length; objNum++) {

			var object = this.objects[objNum];
			if (object['#'] != objNum) {

				window.alert ('Assertion failed');
				return false;
			}

			//------------------------------------------------------
			//	offset for xref
			//------------------------------------------------------

			objOffsets.push(byteOffset);

			//------------------------------------------------------
			//	stream data
			//------------------------------------------------------

			var streamData = null;

			if (object['!']) {

				var joiner = typeof(object['/'])=='string'? object['/'] : '\n';
				streamData = object['!'].join(joiner);
				object.Length = streamData.length;
			}

			//------------------------------------------------------
			//	dictionary
			//------------------------------------------------------

			var builder = [];

			builder.push (objNum + ' 0 obj');

			builder.push ('<<');

			for (name in object) {

				if (name < 'A' || name.substring(0, 1) == '_') continue;
				builder.push('/'+name + ' ' + this.objectToString(object[name], ''));
			}

			builder.push ('>>');

			//------------------------------------------------------
			//	stream data
			//------------------------------------------------------

			if (streamData !== null) {

				builder.push ('stream');
				builder.push (streamData);
				builder.push ('endstream');
			}

			//------------------------------------------------------
			//	end of object
			//------------------------------------------------------

			builder.push ('endobj');
			builder.push ('');
			builder.push ('');
			var string = builder.join('\n');

			components.push(string);
			byteOffset += string.length;
		}

		//--------------------------------------------------------------
		//	XRef
		//--------------------------------------------------------------

		components.push ('xref\n0 ' + this.objects.length + '\n');
		components.push ('0000000000 65535 f\r\n');

		for (objNum=1; objNum<this.objects.length; objNum++) {

			var ofs = this.dl (objOffsets[objNum], 10);
			components.push (ofs + ' 00000 n\r\n');
		}

		//--------------------------------------------------------------
		//	Trailer
		//--------------------------------------------------------------

		var trailer = {
			Info: this.info,
			Root: this.main,
			Size: this.objects.length
		};

		components.push ('trailer\n');
		components.push (this.objectToString(trailer));
		components.push ('\n');

		components.push ('startxref\n' + byteOffset + '\n%%EOF\n');

		return components.join('');
	};
};

//==============================================================================
//	$Id: pdfmaker.js,v 1.32 2019/08/29 15:29:04 wolf Exp $
//==============================================================================
